from peptacular.sequence import get_modifications, add_modifications, strip_modifications, \
    apply_static_modifications, apply_variable_modifications, pop_modifications, calculate_sequence_length, \
    shift_sequence, reverse_sequence, is_sequence_valid, span_to_sequence, split_sequence

import unittest


class TestSequence(unittest.TestCase):

    def test_read_modified_peptide(self):
        # Test the case where there are no modifications in the peptide sequence
        peptide_sequence = "ACDEFG"
        expected_output = {}
        self.assertEqual(get_modifications(peptide_sequence), expected_output)

        # Test the case where there is a single modification in the peptide sequence
        peptide_sequence = "A(2)CDEFG"
        expected_output = {0: 2}
        self.assertEqual(get_modifications(peptide_sequence), expected_output)

        # Test the case where there are multiple modifications in the peptide sequence
        peptide_sequence = "A(2)C(3)DE(1)FG"
        expected_output = {0: 2, 1: 3, 3: 1}
        self.assertEqual(get_modifications(peptide_sequence), expected_output)

    def test_create_modified_peptide(self):
        # Test the case where there are no modifications in the peptide sequence
        unmodified_sequence = "ACDEFG"
        modifications = {}
        expected_output = "ACDEFG"
        self.assertEqual(add_modifications(unmodified_sequence, modifications), expected_output)

        # Test the case where there is a single modification in the peptide sequence
        unmodified_sequence = "ACDEFG"
        modifications = {0: 2}
        expected_output = "A(2)CDEFG"
        self.assertEqual(add_modifications(unmodified_sequence, modifications), expected_output)

        # Test the case where there is a single N-Term mod
        unmodified_sequence = "ACDEFG"
        modifications = {-1: 2}
        expected_output = "[2]ACDEFG"
        self.assertEqual(add_modifications(unmodified_sequence, modifications), expected_output)

        # Test the case where there is a single C-Term mod
        unmodified_sequence = "ACDEFG"
        modifications = {5: 2}
        expected_output = "ACDEFG(2)"
        self.assertEqual(add_modifications(unmodified_sequence, modifications), expected_output)

        # Test the case where there are multiple modifications in the peptide sequence
        unmodified_sequence = "ACDEFG"
        modifications = {0: 2, 1: 3, 3: 1}
        expected_output = "A(2)C(3)DE(1)FG"
        self.assertEqual(add_modifications(unmodified_sequence, modifications), expected_output)

        # Test the case where the index of a modification is invalid for the given peptide sequence
        unmodified_sequence = "ACDEFG"
        modifications = {10: 2}
        with self.assertRaises(ValueError):
            add_modifications(unmodified_sequence, modifications)

    def test_get_unmodified_peptide(self):
        self.assertEqual(strip_modifications('ACDEFG'), 'ACDEFG')
        self.assertEqual(strip_modifications('A(2)CDE(2)FG'), 'ACDEFG')
        self.assertEqual(strip_modifications('A(2)CDE(Acetyl)FG'), 'ACDEFG')

    def test_add_static_mod(self):
        self.assertEqual(apply_static_modifications('PEPCTIDE', {'C': 57.021464}), 'PEPC(57.021464)TIDE')
        self.assertEqual(apply_static_modifications('PEPC(57.021464)TIDE', {'C': 57.021464}), 'PEPC(57.021464)TIDE')
        self.assertEqual(apply_static_modifications('CPEPTIDEC', {'C': 57.021464}), 'C(57.021464)PEPTIDEC(57.021464)')
        self.assertEqual(apply_static_modifications('P(1)EPCTIDE(1)', {'C': 57.021464}), 'P(1)EPC(57.021464)TIDE(1)')

    def test_add_variable_mod(self):
        self.assertEqual(set(apply_variable_modifications('PEPCTIDCE', {'C': 57.021464}, 2)), {'PEPC(57.021464)TIDCE',
                                                                                    'PEPC(57.021464)TIDC(57.021464)E',
                                                                                    'PEPCTIDC(57.021464)E',
                                                                                    'PEPCTIDCE'})

        self.assertEqual(set(apply_variable_modifications('P(1)EPC(20)TIDCE', {'C': 57.021464}, 2)), {'P(1)EPC(20)TIDCE',
                                                                                           'P(1)EPC(20)TIDC(57.021464)E'})

    def test_calculate_sequence_length(self):
        self.assertEqual(calculate_sequence_length("[Acetyl]PEP(1.2345)TID(3.14)E[Amide]"), 7)
        self.assertEqual(calculate_sequence_length("PEPTIDE"), 7)

    def test_get_modifications(self):
        self.assertEqual(get_modifications('PEP(Phospho)T(1)IDE(3.14)'), {2: 'Phospho', 3: 1, 6: 3.14})
        self.assertEqual(get_modifications('[Acetyl]PEPTIDE[Amide]'), {-1: 'Acetyl', 7: 'Amide'})
        self.assertEqual(get_modifications('PEPTIDE'), {})

    def test_add_modifications(self):
        self.assertEqual(add_modifications('PEPTIDE', {2: 'phospho'}), 'PEP(phospho)TIDE')
        self.assertEqual(add_modifications('PEPTIDE', {-1: 'Acetyl', 6: '1.234', 7: 'Amide'}),
                         '[Acetyl]PEPTIDE(1.234)[Amide]')
        self.assertEqual(add_modifications('PEPTIDE', {}), 'PEPTIDE')

    def test_pop_modifications(self):
        self.assertEqual(pop_modifications('PEP(phospho)TIDE'), ('PEPTIDE', {2: 'phospho'}))

    def test_strip_modifications(self):
        self.assertEqual(strip_modifications('PEP(phospho)TIDE'), 'PEPTIDE')

    def test_apply_static_modifications(self):
        self.assertEqual(apply_static_modifications('PEPTIDE', {'P': 'phospho'}), 'P(phospho)EP(phospho)TIDE')

    def test_apply_variable_modifications(self):
        self.assertEqual(apply_variable_modifications('PEPTIDE', {'P': 'phospho'}, 2),
                         ['P(phospho)EP(phospho)TIDE', 'P(phospho)EPTIDE', 'PEP(phospho)TIDE', 'PEPTIDE'])

    def test_reverse_sequence(self):
        self.assertEqual(reverse_sequence('PEPTIDE'), 'EDITPEP')
        self.assertEqual(reverse_sequence('[Acetyl]P(phospho)EP(phospho)TIDE[Amide]'),
                         '[Acetyl]EDITP(phospho)EP(phospho)[Amide]')

    def test_shift_sequence(self):
        self.assertEqual(shift_sequence('PEPTIDE', 2), 'PTIDEPE')

    def test_is_sequence_valid(self):
        self.assertTrue(is_sequence_valid('PEPTIDE'))
        self.assertTrue(is_sequence_valid('[Acetyl]P(phospho)EP(phospho)TIDE[Amide]'))
        self.assertFalse(is_sequence_valid('[Acetyl]P(phospho)EP((phospho))TIDE[Amide]'))

    def test_span_to_sequence(self):
        self.assertEqual(span_to_sequence('PEPTIDE', (0, 4, 0)), 'PEPT')
        self.assertEqual(span_to_sequence('[Acetyl]PEP(1.2345)TID(3.14)E[Amide]', (0, 4, 0)), '[Acetyl]PEP(1.2345)T')

    def test_split_sequence(self):
        self.assertEqual(split_sequence('PEPTIDE'), ['P', 'E', 'P', 'T', 'I', 'D', 'E'])
        self.assertEqual(split_sequence('[Acetyl]P(phospho)EP(phospho)TIDE[Amide]'),
                         ['[Acetyl]P(phospho)', 'E', 'P(phospho)', 'T', 'I', 'D', 'E[Amide]'])


if __name__ == "__main__":
    unittest.main()



import unittest

from peptacular.fragment import Fragment
from peptacular.score import compute_fragment_matches, FragmentMatch, match_spectra_range


class TestScore(unittest.TestCase):

    def test_match_spectra(self):
        # Test basic functionality with 'th' tolerance
        mz_spectrum1 = [100, 200, 300]
        mz_spectrum2 = [99.9, 100.1, 200, 200.05, 300.1, 300.2]
        result = match_spectra_range(mz_spectrum1, mz_spectrum2, 0.2, 'th')
        assert result == [(0, 2), (2, 4), (4, 6)]

        # Test basic functionality with 'th' tolerance
        mz_spectrum1 = [100, 200, 300]
        mz_spectrum2 = [99.9, 100.1, 200, 200.05, 300.1]
        result = match_spectra_range(mz_spectrum1, mz_spectrum2, 0.2, 'th')
        assert result == [(0, 2), (2, 4), (4, 5)]

        # Test with 'ppm' tolerance
        mz_spectrum1 = [100, 200, 300]
        mz_spectrum2 = [99.9, 100.1, 200, 200.05, 300.1, 300.2]
        result = match_spectra_range(mz_spectrum1, mz_spectrum2, 1000, 'ppm')  # 0.1 ppm tolerance
        assert result == [(0, 2), (2, 4), (4, 6)]

        # Test with no matches
        mz_spectrum1 = [50, 150, 250]
        mz_spectrum2 = [100, 200, 300]
        result = match_spectra_range(mz_spectrum1, mz_spectrum2, 0.05, 'th')
        assert result == [None, None, None]

        # Test with empty spectra
        mz_spectrum1 = []
        mz_spectrum2 = [100, 200, 300]
        result = match_spectra_range(mz_spectrum1, mz_spectrum2, 0.1, 'th')
        assert result == []

        mz_spectrum1 = [100, 200, 300]
        mz_spectrum2 = []
        result = match_spectra_range(mz_spectrum1, mz_spectrum2, 0.1, 'th')
        assert result == [None, None, None]



import math
from dataclasses import dataclass
from typing import List, Tuple, Union, Dict
from peptacular.fragment import Fragment


def match_spectra_range(mz_spectrum1: List[float], mz_spectrum2: List[float], tolerance_value: float = 0.1,
                        tolerance_type: str = 'ppm') -> List[Tuple[int, int]]:
    """
    Matches two m/z spectra based on a specified tolerance value and type.

    :param mz_spectrum1: List of m/z values for the first spectrum.
    :type mz_spectrum1: List[float]
    :param mz_spectrum2: List of m/z values for the second spectrum.
    :type mz_spectrum2: List[float]
    :param tolerance_value: Tolerance value for matching. Default is 0.1.
    :type tolerance_value: float
    :param tolerance_type: Type of tolerance ('ppm' or 'th'). Default is 'ppm'.
    :type tolerance_type: str

    :raises ValueError: If the provided tolerance type is not 'ppm' or 'th'.

    :return: List of index pairs representing matched peaks between the two spectra.
    :rtype: List[Tuple[int, int]]

    .. code-block:: python

        >>> match_spectra_range([100, 200, 300], [100, 200, 300])
        [(0, 1), (1, 2), (2, 3)]

        >>> match_spectra_range([100.1, 250, 300, 400], [100, 200, 300], 1, 'th')
        [(0, 1), None, (2, 3), None]

        >>> match_spectra_range([100, 100, 100], [100, 100, 100])
        [(0, 3), (0, 3), (0, 3)]

    """

    if tolerance_type not in ['ppm', 'th']:
        raise ValueError('Invalid tolerance type. Must be "ppm" or "th"')

    indices = []
    mz2_start_index = 0
    for mz1 in mz_spectrum1:

        if mz2_start_index >= len(mz_spectrum2):  # if we've exhausted mz_spectrum2
            indices.append(None)
            continue

        tolerance_offset = tolerance_value if tolerance_type == 'th' else mz1 * tolerance_value / 1e6
        mz1_start = mz1 - tolerance_offset
        mz1_end = mz1 + tolerance_offset

        # Find the first mz2 that is greater than or equal to mz1_start
        while mz2_start_index < len(mz_spectrum2) and mz_spectrum2[mz2_start_index] < mz1_start:
            mz2_start_index += 1

        # If we've reached the end of mz_spectrum2, break
        if mz2_start_index >= len(mz_spectrum2):
            indices.append(None)
            continue

        # Find the last mz2 that is less than or equal to mz1_end
        mz2_end_index = mz2_start_index
        while mz2_end_index < len(mz_spectrum2) and mz_spectrum2[mz2_end_index] <= mz1_end:
            mz2_end_index += 1

        # Decrement to get the actual last index (because while loop exits when condition is no longer met)
        if mz2_end_index - 1 < mz2_start_index:
            indices.append(None)
        else:
            indices.append((mz2_start_index, mz2_end_index))

    return indices


def match_spectra_best(mz_spectrum1: List[float], mz_spectrum2: List[float], tolerance_value: float = 0.1,
                       tolerance_type: str = 'ppm') -> List[Union[int, None]]:
    """
    Matches two m/z spectra based on a specified tolerance value and type.

    :param mz_spectrum1: List of m/z values for the first spectrum.
    :type mz_spectrum1: List[float]
    :param mz_spectrum2: List of m/z values for the second spectrum.
    :type mz_spectrum2: List[float]
    :param tolerance_value: Tolerance value for matching. Default is 0.1.
    :type tolerance_value: float
    :param tolerance_type: Type of tolerance ('ppm' or 'th'). Default is 'ppm'.
    :type tolerance_type: str

    :raises ValueError: If the provided tolerance type is not 'ppm' or 'th'.
    
    :return: List of index pairs representing matched peaks between the two spectra.
    :rtype: List[int]
    
    .. code-block:: python

            >>> match_spectra_best([100, 200, 300], [100, 200, 300])
            [0, 1, 2]

            >>> match_spectra_best([100.1, 250, 300, 400], [100, 200, 300], 1, 'th')
            [0, None, 2, None]

            >>> match_spectra_best([100.1, 250, 300, 400], [100, 100.1, 200, 300], 1, 'th')
            [1, None, 3, None]

    """

    if tolerance_type not in ['ppm', 'th']:
        raise ValueError('Invalid tolerance type. Must be "ppm" or "th"')

    indices = [None] * len(mz_spectrum1)
    j = 0
    for i, mz1 in enumerate(mz_spectrum1):
        tolerance_offset = tolerance_value if tolerance_type == 'th' else mz1 * tolerance_value / 1e6
        mz1_min = mz1 - tolerance_offset
        mz1_max = mz1 + tolerance_offset

        best_match_index = None
        best_match_distance = float('inf')

        while j < len(mz_spectrum2) and mz_spectrum2[j] <= mz1_max:
            if mz1_min <= mz_spectrum2[j]:
                distance = abs(mz1 - mz_spectrum2[j])
                if distance < best_match_distance:
                    best_match_distance = distance
                    best_match_index = j
            j += 1

        indices[i] = best_match_index

    return indices


@dataclass(frozen=True)
class FragmentMatch:
    """
    Represents a match between a theoretical fragment and an experimental spectrum.

    :ivar fragment: Theoretical fragment object.
    :ivar mz: Experimental m/z value.
    :ivar intensity: Intensity of the experimental m/z value.
    """
    fragment: Fragment
    mz: float
    intensity: float

    @property
    def error(self):
        """
        The error between the theoretical and experimental m/z values.

        :return: error between the theoretical and experimental m/z values.
        :rtype: float
        """
        return self.fragment.mz - self.mz

    @property
    def error_ppm(self):
        """
        The error between the theoretical and experimental m/z values in parts-per-million (ppm).

        :return: error between the theoretical and experimental m/z values in parts-per-million (ppm).
        :rtype: float
        """
        return self.error / self.fragment.mz * 1e6


def compute_fragment_matches(fragments: List[Fragment], mz_spectrum: List[float],
                             intensity_spectrum: List[float], tolerance_value: float = 0.1,
                             tolerance_type: str = 'ppm') -> List[FragmentMatch]:
    """
    Computes the fragment matches for a given set of fragments and an experimental spectrum.

    :param fragments: List of theoretical fragments.
    :type fragments: List[Fragment]
    :param mz_spectrum: List of m/z values for the experimental spectrum.
    :type mz_spectrum: List[float]
    :param intensity_spectrum: List of intensity values for the experimental spectrum.
    :type intensity_spectrum: List[float]
    :param tolerance_value: Tolerance value for matching. Default is 0.1.
    :type tolerance_value: float
    :param tolerance_type: Type of tolerance ('ppm' or 'th'). Default is 'ppm'.
    :type tolerance_type: str

    :return: List of fragment matches.
    :rtype: List[FragmentMatch]
    """

    # sort fragments by mass
    fragments.sort(key=lambda x: x.mz)
    fragment_spectrum = [f.mz for f in fragments]
    indices = match_spectra_range(fragment_spectrum, mz_spectrum, tolerance_value, tolerance_type)

    fragment_matches = []
    for i, index in enumerate(indices):
        if index is None:
            continue
        for j in range(*index):
            fragment_matches.append(
                FragmentMatch(fragments[i], mz_spectrum[j], intensity_spectrum[j]))

    return fragment_matches


def hyper_score(fragments: Union[List[Fragment], Dict[str, List[float]]], mz_spectrum: List[float],
                intensity_spectrum: List[float], tolerance_value=0.1, tolerance_type='ppm',
                filter_by='intensity') -> float:
    """
    Computes the hyperscore for a given set of fragments and an experimental spectrum.

    :param fragments: List of theoretical fragments
    :type fragments: List[Fragment] or Dict[str, List[float]]
    :param mz_spectrum: List of m/z values from the experimental spectrum.
    :type mz_spectrum: List[float]
    :param intensity_spectrum: List of intensity values corresponding to the m/z values in mz_spectrum.
    :type intensity_spectrum: List[float]
    :param tolerance_value: Tolerance value for matching fragments to the spectrum.
    :type tolerance_value: float
    :param tolerance_type: Type of tolerance ('ppm' or 'th').
    :type tolerance_type: str
    :param filter_by: How to filter the matched fragments ('intensity' or 'error').
    :type filter_by: str

    :return: Computed hyper score.
    :rtype: float

    .. code-block:: python

            >>> hyper_score({'a': [100, 200, 300]}, [100, 300, 400], [1000, 2000, 1500])
            3.0

    """

    @dataclass(frozen=True)
    class MockFragment:
        mz: float
        ion_type: str

    if isinstance(fragments, Dict):
        fragments = [MockFragment(mz, ion_type) for ion_type in fragments for mz in fragments[ion_type]]

    max_intensity = max(intensity_spectrum)
    intensity_spectrum = [intensity / max_intensity for intensity in intensity_spectrum]
    # Compute the fragment matches using the provided function
    fragment_matches = compute_fragment_matches(fragments, mz_spectrum, intensity_spectrum, tolerance_value,
                                                tolerance_type)

    if len(fragment_matches) == 0:
        return 0

    if filter_by == 'intensity':
        fragment_matches.sort(key=lambda x: x.intensity, reverse=False)
    elif filter_by == 'error':
        fragment_matches.sort(key=lambda x: abs(x.error), reverse=True)

    fragment_matches = {match.fragment: match for match in fragment_matches}.values()

    # Calculate the dot product using matched fragments
    dot_product = sum(match.intensity for match in fragment_matches)

    # Count the number of b and y ions
    Na = sum(1 for match in fragment_matches if match.fragment.ion_type == 'a')
    Nb = sum(1 for match in fragment_matches if match.fragment.ion_type == 'b')
    Nc = sum(1 for match in fragment_matches if match.fragment.ion_type == 'c')
    Nx = sum(1 for match in fragment_matches if match.fragment.ion_type == 'x')
    Ny = sum(1 for match in fragment_matches if match.fragment.ion_type == 'y')
    Nz = sum(1 for match in fragment_matches if match.fragment.ion_type == 'z')

    # Compute the hyper score
    score = dot_product * math.factorial(Na) * math.factorial(Nb) * math.factorial(Nc) * math.factorial(Nx) * \
            math.factorial(Ny) * math.factorial(Nz)

    return score


def _binomial_probability(n: int, k: int, p: float) -> float:
    """
    Computes the binomial probability P(X=k) for given parameters.

    :param n: Number of trials.
    :type n: int
    :param k: Number of successes.
    :type k: int
    :param p: Probability of success in a single trial.
    :type p: int

    :return: Binomial probability P(X=k).
    :rtype: float
    """

    return math.comb(n, k) * (p ** k) * ((1 - p) ** (n - k))


def _estimate_probability_of_random_match(error_tolerance: float, mz_spectrum: List[float],
                                          tolerance_type: str = 'ppm') -> float:
    """
    Estimate the probability of a random match between two peaks based on error tolerance and the experimental spectrum.

    :param error_tolerance: Tolerance value for matching peaks.
    :type error_tolerance: float
    :param mz_spectrum: List of m/z values from the experimental spectrum.
    :type mz_spectrum: List[float]
    :param tolerance_type: Type of tolerance ('ppm' or 'th').
    :type tolerance_type: str

    :return: Estimated probability of a random match.
    :rtype: float
    """

    # Calculate the spectrum range
    spectrum_range = max(mz_spectrum) - min(mz_spectrum)

    # Calculate the error tolerance range
    if tolerance_type == 'ppm':
        average_mz = sum(mz_spectrum) / len(mz_spectrum)
        error_tolerance_range = average_mz * error_tolerance / 1e6
    else:  # 'th'
        error_tolerance_range = error_tolerance

    # Estimate the probability
    num_bins = spectrum_range / error_tolerance_range

    return len(mz_spectrum) / num_bins


def binomial_score(fragments: Union[List[Fragment], List[float]], mz_spectrum: List[float],
                   intensity_spectrum: List[float],
                   tolerance_value=0.1, tolerance_type='ppm') -> float:
    """
    Computes a score based on binomial probability for a given set of fragments and an experimental spectrum.

    :param fragments: List of theoretical fragments.
    :type fragments: List[Fragment]
    :param mz_spectrum: List of m/z values from the experimental spectrum.
    :type mz_spectrum: List[float]
    :param intensity_spectrum: List of intensity values corresponding to the m/z values in mz_spectrum.
    :type intensity_spectrum: List[float]
    :param tolerance_value: Tolerance value for matching fragments to the spectrum.
    :type tolerance_value: float
    :param tolerance_type: Type of tolerance ('ppm' or 'th').
    :type tolerance_type: str

    :return: Score based on binomial probability.
    :rtype: float
    """

    @dataclass(frozen=True)
    class MockFragment:
        mz: float

    if isinstance(fragments, Dict):
        fragments = [MockFragment(mz) for mz in fragments]

    # Compute the fragment matches using the provided function
    fragment_matches = compute_fragment_matches(fragments, mz_spectrum, intensity_spectrum, tolerance_value,
                                                tolerance_type)

    if len(fragment_matches) == 0:
        return 1.0

    fragment_matches = {match.fragment: match for match in fragment_matches}.values()

    # Number of matched fragments (successes)
    k = len(fragment_matches)

    # Total number of trials is the number of theoretical fragments
    n = len(fragments)

    # Estimate the probability of a random match
    p_success = _estimate_probability_of_random_match(tolerance_value, mz_spectrum, tolerance_type)

    # Compute the score based on binomial probability
    score = _binomial_probability(n, k, p_success)

    return score



from typing import Tuple


def strip_n_term_residue(sequence: str) -> str:
    """
    Removes the N-terminal amino acid from the sequence, retaining any other modifications in their original positions.

    If the sequence begins with an N-terminal modification, it will be removed alongside the first amino acid.
    If no modifications are present, this operation is equivalent to eliminating the first character of the sequence.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: Sequence with the N-terminal amino acid (and any N-terminal modification) removed.
    :rtype: str

    .. code-block:: python

        # For sequences without modifications, the first character is removed:
        >>> strip_n_term_residue('PEPTIDE')
        'EPTIDE'

        # Sequences with N-terminal modifications have them removed along with the first amino acid:
        >>> strip_n_term_residue('[Acetyl]P(phospho)EP(phospho)TIDE[Amide]')
        'EP(phospho)TIDE[Amide]'

        # For sequences with a single amino acid having a C-terminal modification, the result is an empty string:
        >>> strip_n_term_residue('E(1)[Amide]')
        ''
        >>> strip_n_term_residue('[Acetyl]P(phospho)[Amide]')
        ''
        >>> strip_n_term_residue('E[Amide]')
        ''

    """

    return sequence[_get_n_term_index(sequence):]


def strip_c_term_residue(sequence: str) -> str:
    """
    Removes the C-terminal amino acid from the sequence, retaining any other modifications in their original positions.

    If the sequence ends with a C-terminal modification, it will be removed alongside the last amino acid.
    If no modifications are present, this operation is equivalent to eliminating the last character of the sequence.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: Sequence with the C-terminal amino acid (and any C-terminal modification) removed.
    :rtype: str

    .. code-block:: python

        # For sequences without modifications, the last character is removed:
        >>> strip_c_term_residue('PEPTIDE')
        'PEPTID'

        # Sequences with C-terminal modifications have them removed along with the last amino acid:
        >>> strip_c_term_residue('[Acetyl]P(phospho)EP(phospho)TIDE(1)[Amide]')
        '[Acetyl]P(phospho)EP(phospho)TID'

        # For sequences with a single amino acid having a N-terminal modification, the result is an empty string:
        >>> strip_c_term_residue('[Acetyl]P(phospho)')
        ''
        >>> strip_c_term_residue('[Acetyl]P(phospho)[Amide]')
        ''
        >>> strip_c_term_residue('[Acetyl]P')
        ''

    """

    return sequence[:_get_c_term_index(sequence)]


def get_n_term_residue(sequence: str) -> str:
    """
    Extracts the N-terminal amino acid from the sequence, along with any  modifications.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: The N-terminal amino acid along with any modifications.
    :rtype: str

    .. code-block:: python

        # For sequences without modifications, the first character is returned:
        >>> get_n_term_residue('PEPTIDE')
        'P'

        # Sequences with N-terminal modifications return them along with the last amino acid:
        >>> get_n_term_residue('[Acetyl]P(phospho)EP(phospho)TIDE(phospho)[Amide]')
        '[Acetyl]P(phospho)'

        # For sequences with a single amino acid
        >>> get_n_term_residue('[Acetyl]P(phospho)[Amide]')
        '[Acetyl]P(phospho)[Amide]'
        >>> get_n_term_residue('[Acetyl]P')
        '[Acetyl]P'
        >>> get_n_term_residue('P[Amide]')
        'P[Amide]'
        >>> get_n_term_residue('P(phospho)[Amide]')
        'P(phospho)[Amide]'
        >>> get_n_term_residue('P(phospho)')
        'P(phospho)'

    """

    end_index = _get_n_term_index(sequence)
    return sequence[:end_index]


def get_c_term_residue(sequence: str) -> str:
    """
    Extracts the C-terminal amino acid from the sequence, along with any modifications.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: The C-terminal amino acid along with any modifications.
    :rtype: str

    .. code-block:: python

        # For sequences without modifications, the last character is returned:
        >>> get_c_term_residue('PEPTIDE')
        'E'

        # Sequences with C-terminal modifications return them along with the last amino acid:
        >>> get_c_term_residue('[Acetyl]P(phospho)EP(phospho)TIDE(phospho)[Amide]')
        'E(phospho)[Amide]'

        # For sequences with a single amino acid
        >>> get_c_term_residue('[Acetyl]P(phospho)[Amide]')
        '[Acetyl]P(phospho)[Amide]'
        >>> get_c_term_residue('[Acetyl]P')
        '[Acetyl]P'
        >>> get_c_term_residue('P[Amide]')
        'P[Amide]'
        >>> get_c_term_residue('P(phospho)[Amide]')
        'P(phospho)[Amide]'
        >>> get_c_term_residue('P(phospho)')
        'P(phospho)'

    """

    start_index = _get_c_term_index(sequence)
    return sequence[start_index:]


def pop_c_term_residue(sequence: str) -> Tuple[str, str]:
    """
    Removes the C-terminal amino acid from the sequence, along with any modifications, and returns it as a tuple with
    the remaining sequence.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: A tuple containing the C-terminal amino acid along with any modifications, and the remaining sequence.
    :rtype: tuple[str, str]

    .. code-block:: python

        # For sequences without modifications, the last character is returned:
        >>> pop_c_term_residue('PEPTIDE')
        ('E', 'PEPTID')

        # Sequences with C-terminal modifications return them along with the last amino acid:
        >>> pop_c_term_residue('[Acetyl]P(phospho)EP(phospho)TIDE(phospho)[Amide]')
        ('E(phospho)[Amide]', '[Acetyl]P(phospho)EP(phospho)TID')

    """

    c_term_index = _get_c_term_index(sequence)
    return sequence[c_term_index:], sequence[:c_term_index]


def pop_n_term_residue(sequence: str) -> Tuple[str, str]:
    """
    Removes the N-terminal amino acid from the sequence, along with any modifications, and returns it as a tuple with
    the remaining sequence.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: A tuple containing the N-terminal amino acid along with any modifications, and the remaining sequence.
    :rtype: tuple[str, str]

    .. code-block:: python

        # For sequences without modifications, the first character is returned:
        >>> pop_n_term_residue('PEPTIDE')
        ('P', 'EPTIDE')

        # Sequences with N-terminal modifications return them along with the last amino acid:
        >>> pop_n_term_residue('[Acetyl]P(phospho)EP(phospho)TIDE(phospho)[Amide]')
        ('[Acetyl]P(phospho)', 'EP(phospho)TIDE(phospho)[Amide]')

    """

    n_term_index = _get_n_term_index(sequence)
    return sequence[:n_term_index], sequence[n_term_index:]


def _get_c_term_index(sequence: str) -> int:
    """
    Returns the start index of the C-terminal amino acid in the sequence.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: The start index of the C-terminal amino acid in the sequence.
    :rtype: int

    .. code-block:: python

        >>> _get_c_term_index('PEPTIDE')
        6
        >>> 'PEPTIDE'[:6]
        'PEPTID'

        >>> _get_c_term_index('[Acetyl]P(phospho)EP(phospho)TIDE(phospho)[Amide]')
        32
        >>> '[Acetyl]P(phospho)EP(phospho)TIDE(phospho)[Amide]'[:32]
        '[Acetyl]P(phospho)EP(phospho)TID'

    """

    start_index = len(sequence) - 1
    if sequence and sequence[start_index] == ']':
        start_index = sequence.rindex('[') - 1

    if sequence[start_index] == ')':
        start_index = sequence.rindex('(') - 1

    # Check for N-Term modification
    if start_index != 0 and sequence[start_index - 1] == ']':
        start_index = 0

    return start_index


def _get_n_term_index(sequence: str) -> int:
    """
    Returns the end index of the N-terminal amino acid in the sequence.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: The end index of the N-terminal amino acid in the sequence.
    :rtype: int

    .. code-block:: python

        >>> _get_n_term_index('PEPTIDE')
        1
        >>> 'PEPTIDE'[1:]
        'EPTIDE'

        >>> _get_n_term_index('[Acetyl]P(phospho)EP(phospho)TIDE(phospho)[Amide]')
        18
        >>> '[Acetyl]P(phospho)EP(phospho)TIDE(phospho)[Amide]'[18:]
        'EP(phospho)TIDE(phospho)[Amide]'

    """

    end_index = 0
    if sequence and sequence[end_index] == '[':
        end_index = sequence.index(']') + 1

    end_index += 1

    if end_index != len(sequence) and sequence[end_index] == '(':
        end_index = sequence.index(')') + 1

    # Check for N-Term modification
    if end_index != len(sequence) and sequence[end_index] == '[':
        end_index = len(sequence)

    return end_index



"""
The `term.py` module provides utilities for handling terminal modifications in amino acid sequences.
These modifications are represented by square brackets either at the N-terminus or the C-terminus.

Key Features:
    - Extract, strip, add, or condense modifications at the N or C terminus.
    - Support for string, float, and integer type modifications.
    - Index-based representation for easy reference.

Term Modification Notation:
    - N-Terminus: [Acetyl]PEPTIDE or [1.234]PEPTIDE
    - C-Terminus: PEPTIDE[Amide] or PEPTIDE[3.1415]
    - N-Terminus modifications are denoted with a -1 index.
    - C-Terminus modifications use the index based on the length of the unmodified sequence.
"""


from typing import Union, Any

from peptacular.util import convert_type


def get_n_term_modification(sequence: str) -> Union[str, float, int, None]:
    """
    Extracts the N-terminal modification from the sequence.

    Given a sequence with potential N-terminal notation, this function will parse and return the N-terminal
    modification, if available. If no modification is present, it returns None.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: N-terminal modification if present, else None.
    :rtype: Union[str, float, int, None]

    .. code-block:: python

        # For string-based modifications:
        >>> get_n_term_modification("[Acetyl]PEPTIDE")
        'Acetyl'

        # For float-based modifications:
        >>> get_n_term_modification("[3.1415]PEPTIDE")
        3.1415

        # For int-based modifications:
        >>> get_n_term_modification("[100]PEPTIDE")
        100

        # When no N-Terminus modification is present:
        >>> get_n_term_modification("PEPTIDE") # returns None

    """

    if sequence.startswith('['):
        # find end notation
        end = sequence.find(']')
        return convert_type(sequence[1:end])
    return None


def get_c_term_modification(sequence: str) -> Union[str, float, int, None]:
    """
    Extracts the C-terminal modification from the provided sequence.

    This function parses a sequence and retrieves the C-terminal modification if it exists.
    If the sequence lacks a C-terminal modification, the function returns None.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: C-terminal modification if present, else None.
    :rtype: Union[str, float, int, None]

    .. code-block:: python

        # For sequences with string-based modifications:
        >>> get_c_term_modification("PEPTIDE[Amide]")
        'Amide'

        # For sequences with float-based modifications:
        >>> get_c_term_modification("PEPTIDE[3.1415]")
        3.1415

        # For sequences with int-based modifications:
        >>> get_c_term_modification("PEPTIDE[100]")
        100

        # When no C-terminal modification is present:
        >>> get_c_term_modification("PEPTIDE")

    """

    if sequence.endswith(']'):
        # find start notation from back
        start = sequence.rfind('[')
        return convert_type(sequence[start + 1:-1])

    return None


def strip_n_term_modification(sequence: str) -> str:
    """
    Removes any N-terminal modification notation from the sequence.

    This function takes a sequence as input and returns the sequence without its N-terminal modification
    notation. If no N-terminal modification exists, the original sequence is returned unchanged.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: Sequence devoid of the N-terminal modification notation.
    :rtype: str

    .. code-block:: python

        # For sequences with different modification types (string, float, int):
        >>> strip_n_term_modification("[Acetyl]PEPTIDE")
        'PEPTIDE'
        >>> strip_n_term_modification("[3.1415]PEPTIDE")
        'PEPTIDE'
        >>> strip_n_term_modification("[100]PEPTIDE")
        'PEPTIDE'

        # If a residue modification is present at the N-terminus, only the terminal notation is removed:
        >>> strip_n_term_modification("[Acetyl]P(1)EPTIDE")
        'P(1)EPTIDE'

        # For sequences without any N-terminal modification:
        >>> strip_n_term_modification("PEPTIDE")
        'PEPTIDE'

    """

    if sequence.startswith('['):
        # find end notation
        end = sequence.find(']')
        return sequence[end + 1:]
    return sequence


def strip_c_term_modification(sequence: str) -> str:
    """
    Removes the C-terminal modification notation from the provided sequence.

    This function parses the input sequence and returns the sequence devoid of
    its C-terminal modification notation. If no C-terminal modification is present in
    the sequence, the original sequence is returned unchanged.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: Sequence without the C-terminal modification notation.
    :rtype: str

    .. code-block:: python

        # For sequences with different modification types (string, float, int):
        >>> strip_c_term_modification("PEPTIDE[Acetyl]")
        'PEPTIDE'
        >>> strip_c_term_modification("PEPTIDE[3.1415]")
        'PEPTIDE'
        >>> strip_c_term_modification("PEPTIDE[100]")
        'PEPTIDE'

        # If a residue modification is present at the C-terminus, only the terminal notation is removed:
        >>> strip_c_term_modification("P(1)EPTIDE[Amide]")
        'P(1)EPTIDE'

        # For sequences without any C-terminal modification:
        >>> strip_c_term_modification("PEPTIDE")
        'PEPTIDE'

    """

    if sequence.endswith(']'):
        # find start notation
        start = sequence.rfind('[')
        return sequence[:start]
    return sequence


def add_n_term_modification(sequence: str, mod: Any) -> str:
    """
    Appends the specified N-terminal modification to the provided sequence.

    If the sequence already contains an N-terminal modification, the new modification will be combined with the
    existing one. Ensure that the types of the modifications are compatible to prevent errors.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param mod: Modification to be appended at the N-terminus of the sequence.
    :type mod: Any

    :return: Modified sequence with the added N-terminal notation.
    :rtype: str

    .. code-block:: python

        # Adding string-based modifications:
        >>> add_n_term_modification("PEPTIDE", "Acetyl")
        '[Acetyl]PEPTIDE'

        # Adding float-based modifications:
        >>> add_n_term_modification("PEPTIDE", 3.1415)
        '[3.1415]PEPTIDE'

        # Adding int-based modifications:
        >>> add_n_term_modification("PEPTIDE", 100)
        '[100]PEPTIDE'

        # Combining new modifications with existing N-terminal modifications:
        >>> add_n_term_modification("[Acetyl]PEPTIDE", "Amide")
        '[AcetylAmide]PEPTIDE'
        >>> add_n_term_modification("[3.1415]PEPTIDE", 3.1415)
        '[6.283]PEPTIDE'
        >>> add_n_term_modification("[100]PEPTIDE", 100)
        '[200]PEPTIDE'

        # In case of incompatible modification types:
        >>> add_n_term_modification("[100]PEPTIDE", '100')
        Traceback (most recent call last):
            ...
        TypeError: unsupported operand type(s) for +: 'int' and 'str'

        # No change if the modification is None:
        >>> add_n_term_modification("PEPTIDE", None)
        'PEPTIDE'

    """

    if mod is None:
        return sequence

    existing_mod = get_n_term_modification(sequence)
    if existing_mod is not None:
        new_mod = existing_mod + mod
        return f"[{str(new_mod)}]{strip_n_term_modification(sequence)}"

    return f"[{str(mod)}]{sequence}"


def add_c_term_modification(sequence: str, mod: Any) -> str:
    """
    Appends the specified C-terminal modification notation to the provided sequence.

    If the sequence already contains a C-terminal modification, the new modification will be
    combined with the existing one. Ensure that the types of the modifications are compatible
    to prevent errors.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param mod: Modification to be appended at the C-terminus of the sequence. Can be of type str, int, or float.
    :type mod: Any

    :return: Modified sequence with the added C-terminal notation.
    :rtype: str

    .. code-block:: python

        # Adding string-based modifications:
        >>> add_c_term_modification("PEPTIDE", "Amide")
        'PEPTIDE[Amide]'

        # Adding float-based modifications:
        >>> add_c_term_modification("PEPTIDE", 3.1415)
        'PEPTIDE[3.1415]'

        # Adding int-based modifications:
        >>> add_c_term_modification("PEPTIDE", 100)
        'PEPTIDE[100]'

        # Combining new modifications with existing C-terminal modifications:
        >>> add_c_term_modification("PEPTIDE[Amide]", "Acetyl")
        'PEPTIDE[AmideAcetyl]'
        >>> add_c_term_modification("PEPTIDE[3.1415]", 3.1415)
        'PEPTIDE[6.283]'
        >>> add_c_term_modification("PEPTIDE[100]", 100)
        'PEPTIDE[200]'

        # In case of incompatible modification types:
        >>> add_c_term_modification("PEPTIDE[100]", '100')
        Traceback (most recent call last):
            ...
        TypeError: unsupported operand type(s) for +: 'int' and 'str'

        # No change if the modification is None:
        >>> add_c_term_modification("PEPTIDE", None)
        'PEPTIDE'

    """

    if mod is None:
        return sequence

    existing_mod = get_c_term_modification(sequence)
    if existing_mod is not None:
        new_mod = existing_mod + mod
        return f"{strip_c_term_modification(sequence)}[{str(new_mod)}]"

    return f"{sequence}[{str(mod)}]"


def strip_term_modifications(sequence: str) -> str:
    """
    Removes both N-terminal and C-terminal modification notations from the sequence.

    If the sequence contains modifications at both terminals, this function will strip both
    notations. If there are no terminal modifications, the sequence is returned unchanged.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: Sequence without any terminal modification notations.
    :rtype: str

    :Examples:

    .. code-block:: python

        # Removing both N-terminal and C-terminal modifications:
        >>> strip_term_modifications("[Acetyl]PEPTIDE[Amide]")
        'PEPTIDE'

        # The result is equivalent to applying strip_n_term_modification followed by strip_c_term_modification:
        >>> strip_n_term_modification(strip_c_term_modification("[Acetyl]PEPTIDE[Amide]"))
        'PEPTIDE'

    """

    return strip_n_term_modification(strip_c_term_modification(sequence))


def condense_n_term_modifications(sequence: str) -> str:
    """
    Merges any N-terminal modification with the modification of the first amino acid.

    If the sequence has both an N-terminal modification and a modification on the first amino acid,
    this function will combine them into a single modification on the first amino acid.
    If there's only an N-terminal modification, it will be transferred to the first amino acid.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: Sequence with merged N-terminal and first amino acid modifications.
    :rtype: str

    .. code-block:: python

        # For sequences with just an N-terminal modification:
        >>> condense_n_term_modifications('[Acetyl]PEPTIDE')
        'P(Acetyl)EPTIDE'

        # Combining N-terminal and first amino acid modifications:
        >>> condense_n_term_modifications('[1.0]P(2)EPTIDE')
        'P(3.0)EPTIDE'
        >>> condense_n_term_modifications('[Acetyl]P(Amide)EPTIDE')
        'P(AcetylAmide)EPTIDE'

        # TypeError when incompatible modifications are attempted to be combined:
        >>> condense_n_term_modifications('[Acetyl]P(1.0)EPTIDE')
        Traceback (most recent call last):
        ...
        TypeError: can only concatenate str (not "float") to str

    """

    n_term_mod = get_n_term_modification(sequence)
    if n_term_mod is None:
        return sequence

    sequence = strip_n_term_modification(sequence)
    if sequence[1] == '(':  # if the first amino acid is modified
        end = sequence.index(')')
        condensed_mod = n_term_mod + convert_type(sequence[2:end])
        sequence = f"{sequence[:1]}({condensed_mod}){sequence[end + 1:]}"
    else:
        sequence = f"{sequence[0]}({n_term_mod}){sequence[1:]}"

    return sequence


def condense_c_term_modifications(sequence: str) -> str:
    """
    Merges any C-terminal modification with the modification of the last amino acid.

    If the sequence has both a C-terminal modification and a modification on the last amino acid,
    this function will combine them into a single modification on the last amino acid.
    If there's only a C-terminal modification, it will be transferred to the last amino acid.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: Sequence with merged C-terminal and last amino acid modifications.
    :rtype: str

    .. code-block:: python

        # For sequences with just a C-terminal modification:
        >>> condense_c_term_modifications('PEPTIDE[Amide]')
        'PEPTIDE(Amide)'

        # Combining C-terminal and last amino acid modifications:
        >>> condense_c_term_modifications('PEPTIDE(1.0)[2.0]')
        'PEPTIDE(3.0)'
        >>> condense_c_term_modifications('PEPTIDE(Acetyl)[Amide]')
        'PEPTIDE(AcetylAmide)'

        # TypeError when incompatible modifications are attempted to be combined:
        >>> condense_c_term_modifications('PEPTIDE(1.0)[Amide]')
        Traceback (most recent call last):
        ...
        TypeError: unsupported operand type(s) for +: 'float' and 'str'

    """

    c_term_mod = get_c_term_modification(sequence)
    if c_term_mod is None:
        return sequence

    sequence = strip_c_term_modification(sequence)
    if sequence[-1] == ')':  # if the last amino acid is modified
        start = sequence.rindex('(')
        condensed_mod = convert_type(sequence[start + 1:-1]) + c_term_mod
        sequence = f"{sequence[:start]}({condensed_mod})"
    else:
        sequence = f"{sequence}({c_term_mod})"

    return sequence


def condense_term_modifications(sequence: str) -> str:
    """
    Merges both the N-terminal and C-terminal modifications with their respective adjacent amino acid modifications.

    If the sequence has terminal modifications (either N-terminal or C-terminal) and adjacent amino acid modifications,
    this function will combine them into single modifications on the respective amino acids.
    If there's only a terminal modification, it will be transferred to the adjacent amino acid.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: Sequence with merged terminal and adjacent amino acid modifications.
    :rtype: str

    .. code-block:: python

        # For sequences with modifications at both terminals:
        >>> condense_term_modifications('[1.0]P(2.0)EPTIDE(1.0)[2.0]')
        'P(3.0)EPTIDE(3.0)'

        # Condense_terms is equivalent to the sequential application of condense_n_term and condense_c_term:
        >>> condense_n_term_modifications(condense_c_term_modifications('[1.0]P(2.0)EPTIDE(1.0)[2.0]'))
        'P(3.0)EPTIDE(3.0)'

    """

    return condense_n_term_modifications(condense_c_term_modifications(sequence))



"""
fragment.py contains functions for.... guess what... fragmenting amino acid sequences!

There are two main methods for generating fragment ions within this module:
- build_fragments() - Generates all fragment objects.
- fragment() - Generates all fragment ion mass values.

fragment() is the recommended way to generate fragment ions, since it is faster and requires less memory than
build_fragments(). Still build_fragments is useful if you want to generate Fragment objects, which contain much more
information about the fragment ions.
"""

from dataclasses import dataclass
from functools import lru_cache, cached_property
from itertools import chain
from typing import List, Generator, Union, Callable

from peptacular.constants import PROTON_MASS, ION_ADJUSTMENTS
from peptacular.mass import calculate_mz, calculate_mass
from peptacular.sequence import calculate_sequence_length, split_sequence
from peptacular.term.residue import strip_c_term_residue, strip_n_term_residue
from peptacular.util import is_forward


@dataclass(frozen=True)
class Fragment:
    """
    A dataclass for representing a peptide fragment ion.

    :ivar str sequence: The fragment sequence, including any modifications.
    :ivar float mz: The fragment mass-to-charge (m/z) ratio
    :ivar int charge: The fragment charge state.
    :ivar str ion_type: The fragment ion type.
    :ivar int number: The fragment ion number, for abc ions this refers to the number of residues starting from the
                      N-terminus, for xyz ions this refers to the number of residues starting from the C-terminus.
    :ivar bool internal: True if the fragment is an internal fragment, False if it is a terminal fragment.
    :ivar int parent_number: The number of the parent fragment from which the internal fragment is derived. For
                             terminal fragments (internal=False) the number and parent number should always be equal.

    .. code-block:: python

        >>> frag = Fragment("PE(3.1415)PTIDE", 1, "b", 2, False, 2, True)
        >>> frag.sequence
        'PE(3.1415)PTIDE'
        >>> frag.charge
        1
        >>> frag.ion_type
        'b'
        >>> frag.number
        2
        >>> frag.internal
        False
        >>> frag.parent_number
        2
        >>> frag.monoisotopic
        True

    """

    sequence: str
    charge: int
    ion_type: str
    number: int
    internal: bool
    parent_number: int
    monoisotopic: bool

    def __post_init__(self):
        """
        Validates the fragment after initialization.
        """

        if self.internal is False and self.number != self.parent_number:
            raise ValueError("Parent number must be equal to number for terminal fragments.")

    @cached_property
    def start(self):
        """
        Returns the start index of the fragment sequence in the parent sequence.

        :return: Start index of the fragment.
        :rtype: int

        :raises ValueError: If the ion type is invalid.
        """

        if self.ion_type in 'abc':
            if self.internal is False:
                return 0
            return self.parent_number - self.number

        if self.ion_type in 'xyz':
            return -self.parent_number

        raise ValueError(f"Invalid ion type: {self.ion_type}")

    @cached_property
    def end(self):
        """
        Returns the end index of the fragment sequence in the parent sequence.

        :return: End index of the fragment.
        :rtype: int

        :raises ValueError: If the ion type is invalid.
        """

        if self.ion_type in 'abc':
            return self.parent_number

        if self.ion_type in 'xyz':
            if self.internal is False:
                return None
            return self.start + self.number

        raise ValueError(f"Invalid ion type: {self.ion_type}")

    @cached_property
    def mass(self):
        """
        Returns the mass of the fragment.

        :return: Mass of the fragment.
        :rtype: float
        """

        return calculate_mz(self.sequence, self.charge, self.ion_type, self.monoisotopic)

    @cached_property
    def neutral_mass(self):
        """
        Returns the neutral mass of the fragment.

        :return: Neutral mass of the fragment.
        :rtype: float
        """

        return calculate_mass(self.sequence, 0, self.ion_type, self.monoisotopic)

    @cached_property
    def mz(self):
        """
        Returns the mass-to-charge (m/z) ratio of the fragment.

        :return: Mass-to-charge (m/z) ratio of the fragment.
        :rtype: float
        """

        return calculate_mz(self.sequence, self.charge, self.ion_type, self.monoisotopic)

    @cached_property
    def label(self):
        """
        Returns the label of the fragment, e.g., b2, y3i, etc.

        :return: Label of the fragment.
        :rtype: str
        """

        return f"{'+' * self.charge}" \
               f"{self.ion_type}" \
               f"{self.parent_number}" \
               f"{'i' if self.internal else ''}" \
               f"{self.number if self.internal else ''}"


def build_fragments(sequence: str, ion_types: Union[List[str], str], charges: Union[List[int], int],
                    monoisotopic: bool = True, internal: bool = False) -> List[Fragment]:
    """
    Builds all Fragment objects or a given input 'sequence'.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param ion_types: A list of ion types to consider, e.g., ['b', 'y'], or a single ion type, e.g., 'b'.
    :type ion_types: Union[List[str], str]
    :param charges: A list of charge states for the fragments, or a single charge state.
    :type charges: Union[List[int], int]
    :param monoisotopic: If True, use monoisotopic masses. If False, use average masses, default is [True].
    :type monoisotopic: bool
    :param internal: If True, include internal fragments. If False, only include terminal fragments, default is [False].
    :type internal: bool

    :return: List of Fragment objects.
    :rtype: List[Fragment]

    .. code-block:: python

        # Create a list of y ions for a modified peptide.
        >>> frags = build_fragments("[1.0]P(2.0)E(3.0)[4.0]", 'y', 1, monoisotopic=True, internal=False)
        >>> len(frags)
        2

        >>> frags[0].sequence
        '[1.0]P(2.0)E(3.0)[4.0]'

        >>> frags[1].sequence
        'E(3.0)[4.0]'

    """

    if isinstance(ion_types, str):
        ion_types = [ion_types]

    if isinstance(charges, int):
        charges = [charges]

    def _fragment(t: str, c: int):
        # Loop through the sequence to generate fragments and yield m/z
        seq_length = calculate_sequence_length(sequence)
        for number, frag in enumerate(build_fragment_sequences(sequence, t)):
            frag_number = seq_length - number
            yield Fragment(sequence=frag, charge=c, ion_type=t, number=frag_number,
                           internal=False, parent_number=frag_number, monoisotopic=monoisotopic)

            if internal is True:
                fragment_seq_len = calculate_sequence_length(frag)
                for internal_number, internal_frag in \
                        enumerate(build_fragment_sequences(sequence=frag, ion_type=t, internal=True)):

                    if internal_number == 0:  # skip first since its already created
                        continue

                    internal_frag_number = fragment_seq_len - internal_number
                    yield Fragment(sequence=internal_frag, charge=c, ion_type=t,
                                   number=internal_frag_number, internal=True, parent_number=frag_number,
                                   monoisotopic=monoisotopic)

    fragments = list(chain.from_iterable(_fragment(t, c) for t in ion_types for c in charges))
    return fragments


def build_fragment_sequences(sequence: str, ion_type: str, internal: bool = False) -> List[str]:
    """
    Creates a list of fragment sequences for a given input `sequence` and `ion_type`.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param ion_type: The ion type, e.g., 'b', 'y', 'a', 'c', 'x', or 'z'.
    :type ion_type: str
    :param internal: If True, build internal fragments. If False, build normal fragments, default is [False].
    :type internal: bool

    :return: List of fragment sequences.
    :rtype: List[str]

    .. code-block:: python

        >>> build_fragment_sequences("PEPTIDE", 'y')
        ['PEPTIDE', 'EPTIDE', 'PTIDE', 'TIDE', 'IDE', 'DE', 'E']

        >>> build_fragment_sequences("PEPTIDE", 'b', True)
        ['PEPTIDE', 'EPTIDE', 'PTIDE', 'TIDE', 'IDE', 'DE', 'E']

        >>> build_fragment_sequences("PEPTIDE", 'b')
        ['PEPTIDE', 'PEPTID', 'PEPTI', 'PEPT', 'PEP', 'PE', 'P']

        >>> build_fragment_sequences("PEPTIDE", 'y', True)
        ['PEPTIDE', 'PEPTID', 'PEPTI', 'PEPT', 'PEP', 'PE', 'P']

    """
    if internal is True:
        return sequence_trimmer(sequence, forward=not is_forward(ion_type))
    return sequence_trimmer(sequence, forward=is_forward(ion_type))


@lru_cache(maxsize=2)
def sequence_trimmer(sequence: str, forward: bool) -> List[str]:
    """
    A generator that yields all subsequences of the input `sequence`, by trimming from the front or back.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param forward: If True, generate sequences from the front. If False, generate sequences from the back.
    :type forward: bool

    :return: The trimmed sequeunces.
    :rtype: List[str]

    .. code-block:: python

        >>> sequence_trimmer("PEPTIDE", True)
        ['PEPTIDE', 'EPTIDE', 'PTIDE', 'TIDE', 'IDE', 'DE', 'E']

        >>> sequence_trimmer("PEPTIDE", False)
        ['PEPTIDE', 'PEPTID', 'PEPTI', 'PEPT', 'PEP', 'PE', 'P']

        # Works with modifications
        >>> sequence_trimmer("[Acetyl]PE(3.0)PTIDE", True)
        ['[Acetyl]PE(3.0)PTIDE', 'E(3.0)PTIDE', 'PTIDE', 'TIDE', 'IDE', 'DE', 'E']

        >>> sequence_trimmer("PEP(1.0)TIDE[Amide]", False)
        ['PEP(1.0)TIDE[Amide]', 'PEP(1.0)TID', 'PEP(1.0)TI', 'PEP(1.0)T', 'PEP(1.0)', 'PE', 'P']

    """

    def trim_from_end(seq: str) -> Generator[str, None, None]:
        while seq:
            yield seq
            seq = strip_c_term_residue(seq)

    def trim_from_start(seq: str) -> Generator[str, None, None]:
        while seq:
            yield seq
            seq = strip_n_term_residue(seq)

    if forward is True:
        return list(trim_from_start(sequence))

    return list(trim_from_end(sequence))


def fragment(sequence: str, ion_types: Union[str, List[str]] = 'y', charges: Union[int, List[int]] = 1,
             monoisotopic: bool = True, internal: bool = False, mz: bool = True) -> List[float]:
    """
    Generates fragment ion mz values.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param ion_types: A list of ion types to consider, e.g., ['b', 'y'], or a single ion type, e.g., 'b'.
    :type ion_types: Union[List[str], str]
    :param charges: A list of charge states for the fragments, or a single charge state.
    :type charges: Union[List[int], int]
    :param monoisotopic: Indicates whether to calculate monoisotopic mass, defaults to [True].
    :type monoisotopic: bool
    :param internal: Indicates whether to calculate internal fragments, defaults to [False].
    :type internal: bool
    :param mz: Indicates whether to return m/z values or mass values, defaults to [True].
    :type mz: bool

    :raise ValueError: If the ion type is not one of 'a', 'b', 'c', 'x', 'y', or 'z'.

    :return: The fragemnt ion m/z values.
    :rtype: List[float]

    .. code-block:: python

        # Fragments are returned from largest to smallest.
        >>> fragment("TIDE", ion_types="y", charges=1)
        [477.21911970781, 376.17144123940005, 263.08737726227, 148.06043423844]
        >>> fragment("TIDE", ion_types="b", charges=2)
        [230.10791574544, 165.58661920145502, 108.07314768954, 51.531115700975]

        # When using multiple ion types and charge states the fragments will be generated first by ion_types
        # and second by charge state. For example, the following will generate the following fragments:
        # [y2+, y1+, y2++, y1++, b2+, b1+, b2++, b1++]
        >>> fragment("PE", ion_types=["y", "b"], charges=[1,2])[:4] # First 4 fragments
        [245.11319808729002, 148.06043423844, 123.06023727703, 74.53385535260499]
        >>> fragment("PE", ion_types=["y", "b"], charges=[1,2])[4:] # Last 4 fragments
        [227.10263340359, 98.06004031562, 114.05495493517999, 49.533658391195004]

        # Can also include modifications in the sequence.
        >>> fragment("[1.0]T(-1.0)IDE(2.0)[-2.0]", ion_types="y", charges=1)
        [477.21911970781, 376.17144123940005, 263.08737726227, 148.06043423844]

        # If the sequence contains a modification that is not a number, an error will be raised.
        >>> fragment("[Acetyl]TIDE", ion_types="b", charges=2)
        Traceback (most recent call last):
        ...
        ValueError: could not convert string to float: 'Acetyl'

        # Can generate internal fragments too, but they are typically not very useful.
        # order: sequences -> [PEP, PE, P,  PEP,  E,  P], ions -> [y3, y3i, y3i, y2, y2i, y1]
        >>> fragment("PEP", ion_types="y", charges=1, internal=True)
        [342.16596193614004, 245.11319808729002, 116.07060499932, 245.11319808729002, 148.06043423844, 116.07060499932]

    """

    if isinstance(ion_types, str):
        ion_types = [ion_types]

    if isinstance(charges, int):
        charges = [charges]

    if internal is True:
        return _slow_fragment(sequence=sequence, ion_types=ion_types, charges=charges, monoisotopic=monoisotopic,
                              mz=mz, internal=internal)

    return _fast_fragment(sequence=sequence, ion_types=ion_types, charges=charges, monoisotopic=monoisotopic, mz=mz)


def _slow_fragment(sequence: str, ion_types: List[str], charges: List[int],
                   monoisotopic: bool, mz: bool, internal: bool) -> List[float]:
    """
    Original fragment function that is slow, but works with internal fragments.
    """

    def _fragment(t: str, c: int, f: Callable):
        # Loop through the sequence to generate fragments and yield m/z
        for sub_sequence in build_fragment_sequences(sequence, ion_type=t):
            if mz is True:
                yield f(sequence=sub_sequence, charge=c, ion_type=t, monoisotopic=monoisotopic)
            if internal is True:
                for i, internal_sub_sequence in enumerate(
                        build_fragment_sequences(sequence=sub_sequence, ion_type=t, internal=True)):
                    if i == 0:  # skip first since its already created
                        continue
                    yield f(sequence=internal_sub_sequence, charge=c, ion_type=t, monoisotopic=monoisotopic)

    func = calculate_mz if mz else calculate_mass
    fragments = list(chain.from_iterable(_fragment(t, c, func) for t in ion_types for c in charges))

    return fragments


def _fast_fragment(sequence: str, ion_types: List[str], charges: List[int],
                   monoisotopic: bool, mz: bool) -> List[float]:
    """
    Faster version of _slow_fragment that does not support internal fragments.
    """
    aas = split_sequence(sequence)

    masses = [calculate_mass(aa, 0, 'b', monoisotopic) for aa in aas]

    fragments = []

    for ion_type in ion_types:
        for charge in charges:

            if ion_type in ['a', 'b', 'c']:
                cum_sum = [sum(masses[:i + 1]) for i in range(len(masses))]
            elif ion_type in ['x', 'y', 'z']:
                cum_sum = [sum(masses[-(i + 1):]) for i in range(len(masses))]
            else:
                raise ValueError(f"Invalid ion type: {ion_type}")

            cum_sum = [x + charge * PROTON_MASS + ION_ADJUSTMENTS[ion_type] for x in cum_sum]

            if mz and charge > 0:
                cum_sum = [x / charge for x in cum_sum]

            fragments.extend(cum_sum[::-1])

    return fragments



import pytest
from peptacular import sequence, digest, fragment, mass, protein, score, spans
from peptacular.term import modification, residue

modules = [
    sequence,
    digest,
    fragment,
    mass,
    protein,
    score,
    spans,
    modification,
    residue
]


@pytest.mark.parametrize("module", modules)
def test_doctests(module):
    import doctest
    doctest.testmod(module)



"""
Digest.py contains functions to handle all your digestion needs and of course it works with peptacular's standard
modification notation. From generating left/right semi enzymatic sequences to calculating protease cleavage sites
there is likely a function for it here.

Proteases can be a name of a protease in peptacular.constants.PROTEASES, or they can be a custom regex string. When
specifying more than one protease, all cleavage sites will be combined (as if both proteases were present at the
same time)
"""

from typing import Union, List

from peptacular.spans import build_left_semi_spans, build_right_semi_spans, build_non_enzymatic_spans, build_spans
from peptacular.constants import PROTEASES
from peptacular.sequence import strip_modifications, calculate_sequence_length, span_to_sequence, get_modifications, \
    _span_to_sequence_fast
from peptacular.util import identify_regex_indexes


def build_left_semi_sequences(sequence: str, min_len: int = None, max_len: int = None) -> List[str]:
    """
    Builds all left-hand semi-enzymatic subsequences derived from the input `sequence`.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param min_len: Minimum length for the subsequences (inclusive), defaults to [None]. If None, min_len will be
                    equal to 1.
    :type min_len: int
    :param max_len: Maximum length for the subsequences (inclusive). If None, the subsequences will go up to
                    1 - length of the `sequence`, defaults to [None].
    :type max_len: Union[int, None]

    :return: Left-hand semi-enzymatic subsequences.
    :rtype: List[str]

    .. code-block:: python

        # Generates all left-hand semi enzymatic sequences (Returned values does not include input sequence)
        >>> build_left_semi_sequences('PEPTIDE')
        ['PEPTID', 'PEPTI', 'PEPT', 'PEP', 'PE', 'P']

        # For single-letter or empty sequences, the function returns an empty list:
        >>> build_left_semi_sequences('P')
        []
        >>> build_left_semi_sequences('')
        []

        # Modifications are preserved:
        >>> build_left_semi_sequences('[1]P(2)EPTIDE')
        ['[1]P(2)EPTID', '[1]P(2)EPTI', '[1]P(2)EPT', '[1]P(2)EP', '[1]P(2)E', '[1]P(2)']

    """

    spans = build_left_semi_spans((0, calculate_sequence_length(sequence), 0), min_len, max_len)
    return [span_to_sequence(sequence, span) for span in spans]


def build_right_semi_sequences(sequence: str, min_len: int = None, max_len: int = None) -> List[str]:
    """
    Builds all right-hand semi-enzymatic subsequences derived from the input `sequence`.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param min_len: Minimum length for the subsequences (inclusive), default is [None]. If None, min_len will be
                    equal to 1.
    :type min_len: int
    :param max_len: Maximum length for the subsequences (inclusive). If None, the subsequences will go up to
                    1 - length of the `sequence`, defaults to [None].
    :type max_len: Union[int, None]

    :return: Right-hand semi-enzymatic subsequences
    :rtype: List[str]

    .. code-block:: python

        # Generates all right-hand semi enzymatic sequences (Returned values does not include input sequence)
        >>> build_right_semi_sequences('PEPTIDE')
        ['EPTIDE', 'PTIDE', 'TIDE', 'IDE', 'DE', 'E']

        # For single-letter or empty sequences, the function returns an empty list:
        >>> build_right_semi_sequences('P')
        []
        >>> build_right_semi_sequences('')
        []

        # Modifications are preserved:
        >>> build_right_semi_sequences('PEPTIDE(1)[2]')
        ['EPTIDE(1)[2]', 'PTIDE(1)[2]', 'TIDE(1)[2]', 'IDE(1)[2]', 'DE(1)[2]', 'E(1)[2]']

    """

    spans = build_right_semi_spans((0, calculate_sequence_length(sequence), 0), min_len, max_len)
    return [span_to_sequence(sequence, span) for span in spans]


def build_semi_sequences(sequence: str, min_len: int = None, max_len: int = None) -> List[str]:
    """
    Builds allsemi-enzymatic sequences from the given input `sequence`.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param min_len: Minimum length for the subsequences (inclusive), default is [None]. If None, min_len will be
                    equal to 1.
    :type min_len: int
    :param max_len: Maximum length for the subsequences (inclusive). If None, the subsequences will go up  to
                    1 - length of the `sequence`, defaults to [None].
    :type max_len: Union[int, None]

    :return: Semi-enzymatic subsequences.
    :rtype: List[str]

    .. code-block:: python

        # Equivalent to build_left_semi_sequences + build_right_semi_sequences
        >>> res = build_left_semi_sequences('PEPTIDE') + build_right_semi_sequences('PEPTIDE')
        >>> build_semi_sequences('PEPTIDE') == res
        True

    """

    return build_left_semi_sequences(sequence, min_len, max_len) + build_right_semi_sequences(sequence, min_len,
                                                                                              max_len)


def build_non_enzymatic_sequences(sequence: str, min_len: int = None, max_len: int = None) -> List[str]:
    """
    Builds all non-enzymatic sequences from the given input `sequence`.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param min_len: Minimum length for the subsequences (inclusive), default is [None]. If None, min_len will be
                    equal to 1.
    :type min_len: int
    :param max_len: Maximum length for the subsequences (inclusive). If None, the subsequences will go up  to
                    1 - length of the `sequence`, defaults to [None].
    :type max_len: Union[int, None]

    :return: Non-enzymatic subsequences
    :rtype: List[str]

    .. code-block:: python

        # Generates non-enzymatic sequences (Returned values does not include input sequence):
        >>> build_non_enzymatic_sequences('PEP')
        ['P', 'PE', 'E', 'EP', 'P']

        # For single-letter or empty sequences, the function returns an empty list:
        >>> build_non_enzymatic_sequences('P')
        []
        >>> build_non_enzymatic_sequences('')
        []

         # Sequences with modifications are processed preserving those modifications:
        >>> build_non_enzymatic_sequences('[Acetyl]P(1.0)EP(1.0)[Amide]')
        ['[Acetyl]P(1.0)', '[Acetyl]P(1.0)E', 'E', 'EP(1.0)[Amide]', 'P(1.0)[Amide]']

    """

    spans = build_non_enzymatic_spans((0, calculate_sequence_length(sequence), 0), min_len, max_len)
    return [span_to_sequence(sequence, span) for span in spans]


def build_enzymatic_sequences(sequence: str, enzyme_regex: str, missed_cleavages: int, semi: bool = False,
                              min_len: int = None, max_len: int = None) -> List[str]:
    """
    Builds all enzymatic sequences from the given input `sequence`.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param enzyme_regex: Regular expression or key in PROTEASES dictionary defining the enzyme's cleavage rule.
    :type enzyme_regex: str
    :param missed_cleavages: Maximum number of missed cleavages.
    :type missed_cleavages: int
    :param semi: Whether to include semi-enzymatic peptides, defaults to [False].
    :type semi: bool
    :param min_len: Minimum length for the subsequences (inclusive), default is [None]. If None, min_len will be
                    equal to 1.
    :type min_len: int
    :param max_len: Maximum length for the subsequences (inclusive). If None, the subsequences will go up  to
                    1 - length of the `sequence`, defaults to [None].
    :type max_len: Union[int, None]

    :return: Enzymatic subsequences.
    :rtype: List[str]

    .. code-block:: python

        # Can specify the name of a protease in the PROTEASES dictionary:
        >>> build_enzymatic_sequences(sequence='TIDERTIDEKTIDE', enzyme_regex='trypsin/P', missed_cleavages=2)
        ['TIDER', 'TIDERTIDEK', 'TIDERTIDEKTIDE', 'TIDEK', 'TIDEKTIDE', 'TIDE']

        # Or specify a regular expression:
        >>> build_enzymatic_sequences(sequence='TIDERTIDEKTIDE', enzyme_regex='([KR])', missed_cleavages=2)
        ['TIDER', 'TIDERTIDEK', 'TIDERTIDEKTIDE', 'TIDEK', 'TIDEKTIDE', 'TIDE']

        # Filter sequences by max length:
        >>> build_enzymatic_sequences(sequence='TIDERTIDEKTIDE', enzyme_regex='([KR])', missed_cleavages=2, max_len=5)
        ['TIDER', 'TIDEK', 'TIDE']

        # Filter sequences by min length:
        >>> build_enzymatic_sequences(sequence='TIDERTIDEKTIDE', enzyme_regex='([KR])', missed_cleavages=2, min_len=6)
        ['TIDERTIDEK', 'TIDERTIDEKTIDE', 'TIDEKTIDE']

        # Modifications are preserved:
        >>> build_enzymatic_sequences(sequence='[1]TIDERT(1.0)IDEKTIDE[2]', enzyme_regex='([KR])', missed_cleavages=2)
        ['[1]TIDER', '[1]TIDERT(1.0)IDEK', '[1]TIDERT(1.0)IDEKTIDE[2]', 'T(1.0)IDEK', 'T(1.0)IDEKTIDE[2]', 'TIDE[2]']

    """

    cleavage_sites = [0] + identify_cleavage_sites(sequence, enzyme_regex) + [calculate_sequence_length(sequence)]
    spans = build_spans(calculate_sequence_length(sequence), cleavage_sites, missed_cleavages, min_len, max_len, semi)
    return [span_to_sequence(sequence, span) for span in spans]


def identify_cleavage_sites(sequence: str, enzyme_regex: str) -> List[int]:
    """
    Return a list of positions where cleavage occurs in input `sequence` based on the provided enzyme regex.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param enzyme_regex: Regular expression or key in PROTEASES dictionary defining enzyme's cleavage rule.
    :type enzyme_regex: str

    :return: List of positions where cleavage occurs in the sequence.
    :rtype: List[int]

    .. code-block:: python

        # Can use a key in PROTEASES to specify the enzyme_regex:
        >>> identify_cleavage_sites(sequence='TIDERTIDEKTIDE', enzyme_regex='trypsin/P')
        [5, 10]

        # Or specify a regular expression:
        >>> identify_cleavage_sites(sequence='TIDERTIDEKTIDE', enzyme_regex='([KR])')
        [5, 10]

        # No cleavage sites are identified if the enzyme_regex does not match the sequence:
        >>> identify_cleavage_sites(sequence='TIDEPTIDEPTIDE', enzyme_regex='trypsin/P')
        []

        # If the protease cleaves at the N-terminus, the first position is included:
        >>> identify_cleavage_sites(sequence='KPEPTIDEK', enzyme_regex='lys-n')
        [8]

        # Similarly, if the protease cleaves at the C-terminus, the last position is included:
        >>> identify_cleavage_sites(sequence='KPEPTIDEK', enzyme_regex='lys-c')
        [1]

        # Will also work with modified sequences
        >>> identify_cleavage_sites(sequence='[Acetyl]TIDERT(1.0)IDEKTIDE[Amide]', enzyme_regex='trypsin/P')
        [5, 10]

        # Non-specific cleavage sites are also identified
        >>> identify_cleavage_sites(sequence='PEPTIDE', enzyme_regex='non-specific')
        [1, 2, 3, 4, 5, 6]

        >>> identify_cleavage_sites(sequence='PPPP', enzyme_regex='PP')
        [1, 2, 3]

    """

    stripped_sequence = strip_modifications(sequence)
    if stripped_sequence != sequence:
        sequence = stripped_sequence

    if enzyme_regex in PROTEASES:
        enzyme_regex = PROTEASES[enzyme_regex]

    last_index = len(stripped_sequence)
    return [i + 1 for i in identify_regex_indexes(sequence, enzyme_regex) if i < last_index - 1]


def digest(sequence: str, enzyme_regex: Union[List[str], str], missed_cleavages: int = 0,
           semi: bool = False, min_len: int = None, max_len: int = None) -> List[str]:
    """
    Returns a list of digested sequences derived from the input `sequence`.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param enzyme_regex: Regular expression or list of regular expressions representing enzyme's cleavage rules.
    :type enzyme_regex: Union[List[str], str]
    :param missed_cleavages: Maximum number of missed cleavages, defaults to [0].
    :type missed_cleavages: int
    :param semi: Whether to include semi-enzymatic peptides, defaults to [False].
    :type semi: bool
    :param min_len: Minimum length for the subsequences (inclusive), default is [None]. If None, min_len will be
                    equal to 1.
    :type min_len: int
    :param max_len: Maximum length for the subsequences (inclusive). If None, the subsequences will go up  to
                    1 - length of the `sequence`, defaults to [None].
    :type max_len: Union[int, None]

    :return: List of digested peptides.
    :rtype: List[str]

    .. code-block:: python

        # Can use a key in PROTEASES to specify the enzyme_regex:
        >>> digest(sequence='TIDERTIDEKTIDE', enzyme_regex='trypsin/P', missed_cleavages=2)
        ['TIDER', 'TIDERTIDEK', 'TIDERTIDEKTIDE', 'TIDEK', 'TIDEKTIDE', 'TIDE']

        # Or specify a regular expression:
        >>> digest(sequence='TIDERTIDEKTIDE', enzyme_regex='([KR])', missed_cleavages=2)
        ['TIDER', 'TIDERTIDEK', 'TIDERTIDEKTIDE', 'TIDEK', 'TIDEKTIDE', 'TIDE']

        # Filter sequecnes by max length:
        >>> digest(sequence='TIDERTIDEKTIDE', enzyme_regex='([KR])', missed_cleavages=2, max_len=5)
        ['TIDER', 'TIDEK', 'TIDE']

        # Filter sequecnes by min length:
        >>> digest(sequence='TIDERTIDEKTIDE', enzyme_regex='([KR])', missed_cleavages=2, min_len=6)
        ['TIDERTIDEK', 'TIDERTIDEKTIDE', 'TIDEKTIDE']

        # Generate semi-enzymatic sequences:
        >>> digest(sequence='TIDERTIDEK(1)TIDE[2]', enzyme_regex='([KR])', missed_cleavages=1, min_len=9, semi=True)
        ['TIDERTIDEK(1)', 'TIDEK(1)TIDE[2]', 'TIDERTIDE', 'IDERTIDEK(1)']

        # Non-specific cleavage sites are also identified
        >>> digest(sequence='PEPT', enzyme_regex='non-specific')
        ['P', 'PE', 'PEP', 'E', 'EP', 'EPT', 'P', 'PT', 'T']

    """

    if min_len is None:
        min_len = 1
    if max_len is None:
        max_len = calculate_sequence_length(sequence)

    if isinstance(enzyme_regex, str):
        enzyme_regex = [enzyme_regex]

    enzyme_regex = [PROTEASES.get(regex, regex) for regex in enzyme_regex]

    seq_len = calculate_sequence_length(sequence)

    cleavage_sites = []
    for regex in enzyme_regex:
        cleavage_sites.extend(identify_cleavage_sites(sequence, regex))

    stripped_sequence = strip_modifications(sequence)
    mods = get_modifications(sequence)

    spans = build_spans(seq_len, cleavage_sites, missed_cleavages, min_len, max_len, semi)
    sequences = [_span_to_sequence_fast(stripped_sequence, mods, span) for span in spans]

    return sequences



from typing import Union, List, Tuple
import regex as re

from peptacular.constants import VALID_ION_TYPES, FORWARD_IONS


def validate_ion_type(ion_type: str) -> None:
    """
    Validates if the given ion type is valid.

    This function checks if the ion type provided is one of the valid ion types. If not, it raises a ValueError.

    :param ion_type: The type of ion to validate.
    :type ion_type: str

    :raises ValueError: If the ion type is not valid.

    .. code-block:: python

        >>> validate_ion_type("a")  # No error raised

        >>> validate_ion_type("z")  # No error raised

        >>> validate_ion_type("h")  # Raises ValueError
        Traceback (most recent call last):
        ValueError: Ion type h is invalid.
    """

    if ion_type not in VALID_ION_TYPES:
        raise ValueError(f"Ion type {ion_type} is invalid.")


def is_forward(ion_type: str) -> bool:
    """
        Checks if the provided ion type is a forward ion.

        This function first validates the ion type and then checks if it's a forward ion.

        :param ion_type: The type of ion to check.
        :type ion_type: str

        :return: True if it's a forward ion, False otherwise.
        :rtype: bool

        .. code-block:: python

            >>> is_forward("a")  # Returns False
            False

            >>> is_forward("y")  # Returns True
            True

    """

    validate_ion_type(ion_type)
    return ion_type not in FORWARD_IONS


def _are_parentheses_balanced(text: str, open_char='(', closed_char=')') -> bool:
    """
    Check if parentheses in the given text are balanced or not.

    :param text: The input string to check for balanced parentheses.
    :type text: str

    :return: True if parentheses are balanced, False otherwise.
    :rtype: bool

    .. code-block:: python

        >>> _are_parentheses_balanced("PEPTIDE(1.2345)")
        True

        >>> _are_parentheses_balanced("PEPTIDE(1.2345")
        False

        >>> _are_parentheses_balanced("PEPTIDE((1.2345")
        False

        >>> _are_parentheses_balanced("PEPTIDE((1.2345)")
        False

    """
    stack = []
    for i, char in enumerate(text):
        if char == open_char:
            stack.append(i)
        elif char == closed_char:
            if not stack:
                return False
            stack.pop()
    return len(stack) == 0


def validate_parentheses(sequence: str) -> None:
    """
    Validate the parentheses in the given sequence.

    :param sequence: The sequence to validate.
    :type sequence: str

    :raises ValueError: If the parentheses are not balanced.

    .. code-block:: python
        >>> validate_parentheses("PEPTIDE(1.2345)")

        >>> validate_parentheses("PEPTIDE(1.2345")
        Traceback (most recent call last):
        ValueError: Incorrect modification notation in peptide sequence: "PEPTIDE(1.2345".

        >>> validate_parentheses("PEPTIDE(1.2345)[Amide")
        Traceback (most recent call last):
        ValueError: Incorrect modification notation in peptide sequence: "PEPTIDE(1.2345)[Amide".

    """
    # Check parentheses balance
    if not _are_parentheses_balanced(sequence, '(', ')'):
        raise ValueError(f'Incorrect modification notation in peptide sequence: "{sequence}".')

    if not _are_parentheses_balanced(sequence, '[', ']'):
        raise ValueError(f'Incorrect modification notation in peptide sequence: "{sequence}".')


def convert_type(val: str) -> Union[str, int, float]:
    """
    Convert the given value to the appropriate type.

    :param val: The value to convert.
    :type val: str

    :return: The converted value.
    :rtype: Union[str, int, float]

    .. code-block:: python

        >>> convert_type("1.234")
        1.234

        >>> convert_type("1")
        1

        >>> convert_type("abc")
        'abc'
    """
    try:
        return int(val)
    except ValueError:
        try:
            return float(val)
        except ValueError:
            return val


def identify_regex_indexes(input_str: str, regex_str: str, offset: int = 0) -> List[int]:
    """
    Identify the starting indexes of occurrences of a given regex pattern within a string.

    :param input_str: The sequence in which to search.
    :type input_str: str
    :param regex_str: The regex pattern to search for.
    :type regex_str: str
    :param offset: An optional offset to add to each identified index. Default is 0.
    :type offset: int
    :return: A list of starting indexes where the regex pattern is found in the sequence.
    :rtype: List[int]

    .. code-block:: python

        >>> identify_regex_indexes("PEPTIDE", "P")
        [0, 2]

        >>> identify_regex_indexes("PEPTIDE", "E")
        [1, 6]

        # More complex regex
        >>> identify_regex_indexes("PEPTIDEP", "P[ST]")
        [2]

        >>> identify_regex_indexes("PPPP", "PP")
        [0, 1, 2]

    """

    regex_indexes = [match.start() + offset for match in re.finditer(regex_str, input_str, overlapped=True)]
    return regex_indexes


def _validate_span(span: Tuple[int, int, int]) -> None:
    """
    Validates if a given span is valid.

    :param span: A tuple representing the span.
    :type span: Tuple[int, int, int]
    :raises ValueError: If the span is not valid.

    .. code-block:: python

        >>> _validate_span((0, 5, 0))  # No error raised

        >>> _validate_span((0, 0, 0))  # No error raised

        >>> _validate_span((5, 0, 0))  # Raises ValueError
        Traceback (most recent call last):
        ValueError: Start of span: 5, should be less than or equal to end of span: 0.

        >>> _validate_span((-1, 0, 0))  # Raises ValueError
        Traceback (most recent call last):
        ValueError: Start of span should be non-negative, got -1.

        >>> _validate_span((0, -1, 0))  # Raises ValueError
        Traceback (most recent call last):
        ValueError: End of span should be non-negative, got -1.

    """

    start, end, _ = span
    if start < 0:
        raise ValueError(f'Start of span should be non-negative, got {start}.')
    if end < 0:
        raise ValueError(f'End of span should be non-negative, got {end}.')
    if start > end:
        raise ValueError(f'Start of span: {start}, should be less than or equal to end of span: {end}.')



import regex as re
from typing import List


def find_peptide_indexes(protein: str, peptide: str) -> List[int]:
    """
    Retrieves all starting indexes of a given peptide within a protein sequence.

    :param protein: The complete protein sequence in which to search.
    :type protein: str
    :param peptide: The peptide sequence to find within the protein.
    :type peptide: str
    :return: A list of starting indexes where the peptide is found in the protein sequence.
    :rtype: List[int]

    .. code-block:: python

        >>> find_peptide_indexes("PEPTIDE", "PEP")
        [0]

        >>> find_peptide_indexes("PEPTIDE", "EPT")
        [1]

        >>> find_peptide_indexes("PEPTIDE", "E")
        [1, 6]

    """

    if len(peptide) == 0:
        return []

    return [i.start() for i in re.finditer(peptide, protein, overlapped=True)]


def build_coverage_array(protein: str, peptides: List[str], accumulate: bool = False) -> List[int]:
    """
    Calculate the coverage of a protein sequence by a list of peptides.

    The coverage is represented as a binary list where each position in the protein sequence is marked as 1 if it
    is covered by at least one peptide and 0 otherwise.

    :param protein: The protein sequence.
    :type protein: str
    :param peptides: List of peptide sequences.
    :type peptides: List[str]
    :param accumulate: If True, the coverage array will be accumulated, i.e. if a position is covered by more than
                        one peptide, it will be marked as the sum of the number of peptides covering it. If False,
                        the coverage array will be binary, i.e. if a position is covered by more than one peptide,
                        it will be marked as 1.
    :type accumulate: bool

    :return: A list representing the coverage of the protein sequence by the peptides. Each position in the
             list corresponds to a position in the protein sequence.
    :rtype: List[int]

    .. code-block:: python

        >>> build_coverage_array("PEPTIDE", ["PEP"])
        [1, 1, 1, 0, 0, 0, 0]

        >>> build_coverage_array("PEPTIDE", ["PEP", "EPT"])
        [1, 1, 1, 1, 0, 0, 0]

        # If accumulate is True, overlapping indecies will be accumulated
        >>> build_coverage_array("PEPTIDE", ["PEP", "EPT"], accumulate=True)
        [1, 2, 2, 1, 0, 0, 0]

    """

    cov_arr = [0] * len(protein)
    for peptide in peptides:
        peptide_indexes = find_peptide_indexes(protein, peptide)
        for peptide_index in peptide_indexes:
            if accumulate:
                cov_arr[peptide_index:peptide_index + len(peptide)] = \
                    [x + 1 for x in cov_arr[peptide_index:peptide_index + len(peptide)]]
            else:
                cov_arr[peptide_index:peptide_index + len(peptide)] = [1] * len(peptide)

    return cov_arr


def calculate_percent_coverage(protein: str, peptides: List[str]) -> float:
    """
    Calculates the protein coverage of a list of peptides as a percentage.

    :param protein: The protein sequence.
    :type protein: str
    :param peptides: The list of peptide sequences.
    :type peptides: List[str]

    :return: The protein coverage percentage.
    :rtype: float

    .. code-block:: python

        >>> calculate_percent_coverage("PEPTIDE", ["PEP"])
        0.42857142857142855

        >>> calculate_percent_coverage("PEPTIDE", ["PEP", "EPT"])
        0.5714285714285714

    """

    cov_arr = build_coverage_array(protein, peptides, accumulate=False)

    if len(cov_arr) == 0:
        return 0

    return sum(cov_arr) / len(cov_arr)



from peptacular.constants import MONO_ISOTOPIC_ATOMIC_MASSES, AVERAGE_ATOMIC_MASSES
from peptacular.mass import calculate_mz, calculate_mass

import unittest


class TestMass(unittest.TestCase):
    def test_calculate_mz_with_unmodified_peptide(self):
        sequence = 'PEPTIDE'
        places = 2

        self.assertAlmostEqual(799.359964, calculate_mz(sequence, charge=0, ion_type='y', monoisotopic=True, uwpr_mass=True), places)
        self.assertAlmostEqual(800.367241, calculate_mz(sequence, charge=1, ion_type='y', monoisotopic=True, uwpr_mass=True), places)
        self.assertAlmostEqual(400.687259, calculate_mz(sequence, charge=2, ion_type='y', monoisotopic=True, uwpr_mass=True), places)
        self.assertAlmostEqual(267.460598, calculate_mz(sequence, charge=3, ion_type='y', monoisotopic=True, uwpr_mass=True), places)
        self.assertAlmostEqual(200.847268, calculate_mz(sequence, charge=4, ion_type='y', monoisotopic=True, uwpr_mass=True), places)

        # Average mass is off by 0.004 Da
        self.assertAlmostEqual(799.822520, calculate_mz(sequence, charge=0, ion_type='y', monoisotopic=False, uwpr_mass=True), places)
        self.assertAlmostEqual(800.829796, calculate_mz(sequence, charge=1, ion_type='y', monoisotopic=False, uwpr_mass=True), places)
        self.assertAlmostEqual(400.918536, calculate_mz(sequence, charge=2, ion_type='y', monoisotopic=False, uwpr_mass=True), places)
        self.assertAlmostEqual(267.614783, calculate_mz(sequence, charge=3, ion_type='y', monoisotopic=False, uwpr_mass=True), places)
        self.assertAlmostEqual(200.962906, calculate_mz(sequence, charge=4, ion_type='y', monoisotopic=False, uwpr_mass=True), places)

    def test_calculate_mz_with_modified_peptide(self):
        sequence = '(15)P(-10)EPTIDE(100)'
        places = 2

        self.assertAlmostEqual(799.359964 + 105, calculate_mz(sequence, charge=0, ion_type='y', monoisotopic=True, uwpr_mass=True),
                               places)
        self.assertAlmostEqual(800.367241 + 105 / 1, calculate_mz(sequence, charge=1, ion_type='y', monoisotopic=True, uwpr_mass=True),
                               places)
        self.assertAlmostEqual(400.687259 + 105 / 2, calculate_mz(sequence, charge=2, ion_type='y', monoisotopic=True, uwpr_mass=True),
                               places)
        self.assertAlmostEqual(267.460598 + 105 / 3, calculate_mz(sequence, charge=3, ion_type='y', monoisotopic=True, uwpr_mass=True),
                               places)
        self.assertAlmostEqual(200.847268 + 105 / 4, calculate_mz(sequence, charge=4, ion_type='y', monoisotopic=True, uwpr_mass=True),
                               places)

        self.assertAlmostEqual(799.822520 + 105, calculate_mz(sequence, charge=0, ion_type='y', monoisotopic=False, uwpr_mass=True),
                               places)
        self.assertAlmostEqual(800.829796 + 105 / 1, calculate_mz(sequence, charge=1, ion_type='y', monoisotopic=False, uwpr_mass=True),
                               places)
        self.assertAlmostEqual(400.918536 + 105 / 2, calculate_mz(sequence, charge=2, ion_type='y', monoisotopic=False, uwpr_mass=True),
                               places)
        self.assertAlmostEqual(267.614783 + 105 / 3, calculate_mz(sequence, charge=3, ion_type='y', monoisotopic=False, uwpr_mass=True),
                               places)
        self.assertAlmostEqual(200.962906 + 105 / 4, calculate_mz(sequence, charge=4, ion_type='y', monoisotopic=False, uwpr_mass=True),
                               places)

    def test_calculate_mass_with_modified_peptide(self):
        sequence = 'PEPTIDE'
        places = 2

        self.assertAlmostEqual(799.359964, calculate_mass(sequence, charge=0, ion_type='y', monoisotopic=True, uwpr_mass=True), places)
        self.assertAlmostEqual(799.359964 + MONO_ISOTOPIC_ATOMIC_MASSES['PROTON'] * 1,
                               calculate_mass(sequence, charge=1, ion_type='y', monoisotopic=True, uwpr_mass=True), places)
        self.assertAlmostEqual(799.359964 + MONO_ISOTOPIC_ATOMIC_MASSES['PROTON'] * 2,
                               calculate_mass(sequence, charge=2, ion_type='y', monoisotopic=True, uwpr_mass=True), places)
        self.assertAlmostEqual(799.359964 + MONO_ISOTOPIC_ATOMIC_MASSES['PROTON'] * 3,
                               calculate_mass(sequence, charge=3, ion_type='y', monoisotopic=True, uwpr_mass=True), places)
        self.assertAlmostEqual(799.359964 + MONO_ISOTOPIC_ATOMIC_MASSES['PROTON'] * 4,
                               calculate_mass(sequence, charge=4, ion_type='y', monoisotopic=True, uwpr_mass=True), places)

        self.assertAlmostEqual(799.822520, calculate_mass(sequence, charge=0, ion_type='y', monoisotopic=False, uwpr_mass=True), places)
        self.assertAlmostEqual(799.822520 + AVERAGE_ATOMIC_MASSES['PROTON'] * 1,
                               calculate_mass(sequence, charge=1, ion_type='y', monoisotopic=False, uwpr_mass=True), places)
        self.assertAlmostEqual(799.822520 + AVERAGE_ATOMIC_MASSES['PROTON'] * 2,
                               calculate_mass(sequence, charge=2, ion_type='y', monoisotopic=False, uwpr_mass=True), places)
        self.assertAlmostEqual(799.822520 + AVERAGE_ATOMIC_MASSES['PROTON'] * 3,
                               calculate_mass(sequence, charge=3, ion_type='y', monoisotopic=False, uwpr_mass=True), places)
        self.assertAlmostEqual(799.822520 + AVERAGE_ATOMIC_MASSES['PROTON'] * 4,
                               calculate_mass(sequence, charge=4, ion_type='y', monoisotopic=False, uwpr_mass=True), places)

    def test_calculate_mass_with_unmodified_peptide(self):
        sequence = '(15)P(-10)EPTIDE(100)'
        places = 2

        self.assertAlmostEqual(799.359964 + 105, calculate_mass(sequence, charge=0, ion_type='y', monoisotopic=True, uwpr_mass=True),
                               places)
        self.assertAlmostEqual(799.359964 + MONO_ISOTOPIC_ATOMIC_MASSES['PROTON'] * 1 + 105,
                               calculate_mass(sequence, charge=1, ion_type='y', monoisotopic=True, uwpr_mass=True), places)
        self.assertAlmostEqual(799.359964 + MONO_ISOTOPIC_ATOMIC_MASSES['PROTON'] * 2 + 105,
                               calculate_mass(sequence, charge=2, ion_type='y', monoisotopic=True, uwpr_mass=True), places)
        self.assertAlmostEqual(799.359964 + MONO_ISOTOPIC_ATOMIC_MASSES['PROTON'] * 3 + 105,
                               calculate_mass(sequence, charge=3, ion_type='y', monoisotopic=True, uwpr_mass=True), places)
        self.assertAlmostEqual(799.359964 + MONO_ISOTOPIC_ATOMIC_MASSES['PROTON'] * 4 + 105,
                               calculate_mass(sequence, charge=4, ion_type='y', monoisotopic=True, uwpr_mass=True), places)

        self.assertAlmostEqual(799.822520 + 105, calculate_mass(sequence, charge=0, ion_type='y', monoisotopic=False, uwpr_mass=True),
                               places)
        self.assertAlmostEqual(799.822520 + AVERAGE_ATOMIC_MASSES['PROTON'] * 1 + 105,
                               calculate_mass(sequence, charge=1, ion_type='y', monoisotopic=False, uwpr_mass=True), places)
        self.assertAlmostEqual(799.822520 + AVERAGE_ATOMIC_MASSES['PROTON'] * 2 + 105,
                               calculate_mass(sequence, charge=2, ion_type='y', monoisotopic=False, uwpr_mass=True), places)
        self.assertAlmostEqual(799.822520 + AVERAGE_ATOMIC_MASSES['PROTON'] * 3 + 105,
                               calculate_mass(sequence, charge=3, ion_type='y', monoisotopic=False, uwpr_mass=True), places)
        self.assertAlmostEqual(799.822520 + AVERAGE_ATOMIC_MASSES['PROTON'] * 4 + 105,
                               calculate_mass(sequence, charge=4, ion_type='y', monoisotopic=False, uwpr_mass=True), places)
